/* 
 * All modifications in this file to the original code are
 * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
 *
 * for details see file COPYING.DOSEMU in the DOSEMU distribution
 */

int bootoff_main(int argc, char **argv);
int booton_main(int argc, char **argv);
int dpmi_main(int argc, char **argv);
int ecpuoff_main(int argc, char **argv);
int ecpuon_main(int argc, char **argv);
int eject_main(int argc, char **argv);
int emumouse_main(int argc, char **argv);
int exitemu_main(int argc, char **argv);
int speed_main(int argc, char **argv);
int system_main(int argc, char **argv);
int uchdir_main(int argc, char **argv);
int ugetcwd_main(int argc, char **argv);
int vgaon_main(int argc, char **argv);
int vgaoff_main(int argc, char **argv);
