/* 
 * All modifications in this file to the original code are
 * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
 *
 * for details see file COPYING.DOSEMU in the DOSEMU distribution
 */

void make_unmake_dos_mangled_path(char *dest, char *fpath,
                                         int current_drive, int alias);
void close_dirhandles(unsigned psp);
