/*
 * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
 *
 * for details see file COPYING.DOSEMU in the DOSEMU distribution
 */

#ifndef VESA_H
#define VESA_H

extern void vesa_init(void);
extern char *vesa_get_lfb(void);

#endif
