/* 
 * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
 *
 * for details see file COPYING.DOSEMU in the DOSEMU distribution
 */

int  dis_8086(const unsigned char *, unsigned char *, int, unsigned int *, int);


