/* 
 * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
 *
 * for details see file COPYING.DOSEMU in the DOSEMU distribution
 */

void get_VXD_entry( struct sigcontext *scp );
void vxd_call(struct sigcontext *scp);
